#include <curses.h>
#include "snake.h"

snake *make_snake (char *, char, int, int );
void step_snake (snake *);
void draw_snake (snake *);
void snake_info (snake *);
int init_screen (void);
void error (char *);
int parse_input (void);
void eat_tail ( snake * );
void move_head ( snake * );
void set_keys ( snake *, int, int, int, int, int, int );
void toggle_grow ( snake * );
void draw_motion (void);


snake *zoo[MAXPLAYER];		/* snakes pointer */
int delay = 20;			/* game delay in milliseconds */
char sprite[] = { '-','/','|','\\' };
char spriteidx = 0;

void main()
{
	int i;

	/* initialize the screen or exit */
	if ( ! init_screen() ) error("can't initialize screen");

	/* create the snake */
 	zoo[0] = make_snake("mario", '@', (LINES-2)/2, ((COLS-2)/4)*3 );
	set_keys ( zoo[0], 'k', 'o', ';', 'l', 'h', 'j' );

 	zoo[1] = make_snake("mario", '*', (LINES-2)/2, ((COLS-2)/4)*1 );
	set_keys ( zoo[1], 'a', 'w', 'd', 's', 'z', 'x' );

/*
 	zoo[2] = make_snake("mario", '%', 15, 18 );
	set_keys ( zoo[2], '4', '8', '6', '2', 'z', 'x' );
*/

	/* draw the ring */
	box(stdscr,ACS_VLINE,ACS_HLINE);

	/* draw the snakes */
	for ( i=0; zoo[i]; i++ ) draw_snake(zoo[i]);

	/* refresh the screen */
	refresh();

	while ( parse_input() != 0 )
	{
		/* show activity */
		draw_motion();

		/* have the snakes step */
		for ( i=0; zoo[i]; i++ ) step_snake(zoo[i]);

		/* refresh the screen with the new data */
		refresh();
		
		/* sleep for delay milliseconds */
		napms(delay);

	}

	/* reset the screen */
	endwin();

	/* snake_info(zoo[0]); */
}


/* initialize the screen  */
int init_screen ()
{
	/* initialize the COLS and LINES vars */
	initscr();

	/* make cursor invisible */
	if ( curs_set(0) == ERR ) return 0;	

	/* make each carachter break input reads */
	if ( cbreak() == ERR ) return 0;			

	/* make getch() be a non-blocking call  */
	if ( nodelay(stdscr, TRUE) == ERR ) return 0;

	/* redraw completely on refresh */
	clearok(stdscr, TRUE);

	/* success */
	return 1;
}


/* initialize a snake structure allocating memory for its member  */
snake *make_snake ( char *name, char look, int y, int x )
{
	/* new snake pointer */
	snake *new;	

	/* allocate memory for the new snake */
	new = (snake *) malloc(sizeof(snake));

	/* set snake's look */
	new->look = look | A_REVERSE;

	/* set snake's name */
	strncpy(new->name, name, MAXNAMLEN-1);

	/* allocate space for the head/tail */
	new->head = new->tail = (chain *) malloc(sizeof(chain));

	/* set tail to NULL */
	new->head->next = NULL;

	/* set snake's head and tail chains */
	new->head->y = y;	/* = new->tail->y */
	new->head->x = x;	/* = new->tail->x */

	/* give snake a life */
	new->died = FALSE;

	/* mark it growing */
	new->grow = TRUE;

	/* return the new snake */
	return new;
}


/* make the snake take a step */
void step_snake ( snake *s )
{

	/* snake is dead */
	if ( s->died == TRUE ) return;

	/* move_head if not stopped */
	if ( s->dir != STOP ) move_head(s);

	/* eat the tail if not growing */
	if ( s->grow == FALSE ) eat_tail(s);

}
		



/* print snake information */
void snake_info ( snake *s )
{
	int i=0;
	chain *ptr;

	/* snake's name */
	printf("snake's name: %s\n", s->name);

	/* snake's look */
	printf("snake's look: %c\n", s->look);

	/* snake direction */
	printf("snake's direction: %d\n", s->dir);

	/* snake's control keys */
	printf("snake's control keys (left, up, right, down): %c, %c, %c, %c\n", \
		s->kl, s->ku, s->kr, s->kd);

	printf("\n");
}

/* exit on error printing a message */
void error ( char *msg)
{
	printf("%s\n", msg);
	endwin();
	exit(1);
}


/* parse user input */
int parse_input ()
{
	int ch, i;

	/* get a character */
	ch=getch();

	/* return 0 when reading the quit command */
	if ( ch == 'q' ) return 0;

	/* lower delay if '-' */
	if ( ch == '-' ) 
	{
		delay += 5;
		if ( delay > 1000 ) delay = 1000;
		return 1;
	}
	
	/* higher delay if '+'  */
	if ( ch == '+' )
	{
		delay -= 5;
		if ( delay < 10 ) delay = 10;
		return 1;
	}

	/* check for each snake if any valid key is valid */
	for ( i=0; zoo[i]; i++ )
	{
		chain ptr;

		/* go to next iteration if this snake is dead */
		if ( zoo[i]->died ) continue;

		if ( ch == zoo[i]->kl )
		{
			if ( zoo[i]->dir == RIGHT ) zoo[i]->dir = STOP;
			else if ( zoo[i]->dir == LEFT ) toggle_grow(zoo[i]);

			else
			{
				ptr.x = zoo[i]->head->x - 1;
				if ( ptr.x < 0 ) ptr.x = COLS - 1;

				/* toggle growth if trying to go back */
				if ( zoo[i]->head->next && zoo[i]->head->next->x == ptr.x )
					toggle_grow(zoo[i]);

				/* change to left direction */
				else zoo[i]->dir = LEFT;
			}

			continue;
		}

		if ( ch == zoo[i]->ku )
		{
			if ( zoo[i]->dir == DOWN ) zoo[i]->dir = STOP;
			else if ( zoo[i]->dir == UP ) toggle_grow(zoo[i]);

			else
			{
				ptr.y = zoo[i]->head->y - 1;
				if ( ptr.y < 0 ) ptr.y = LINES-1;

				/* toggle growth if trying to go back */
				if ( zoo[i]->head->next && zoo[i]->head->next->y == ptr.y )
						toggle_grow(zoo[i]);

				/* change to right direction */
				else zoo[i]->dir = UP;
			}

			continue;
		}

		if ( ch == zoo[i]->kr )
		{

			if ( zoo[i]->dir == LEFT ) zoo[i]->dir = STOP;

			else if ( zoo[i]->dir == RIGHT ) toggle_grow(zoo[i]);

			else
			{
				ptr.x = zoo[i]->head->x + 1;
				if ( ptr.x == COLS ) ptr.x = 0;

				/* toggle growth if trying to go back */
				if ( zoo[i]->head->next && zoo[i]->head->next->x == ptr.x )
					toggle_grow(zoo[i]);

				/* change to right direction */
				else zoo[i]->dir = RIGHT;
			}

			continue;

		}


		if ( ch == zoo[i]->kd )
		{

			if ( zoo[i]->dir == UP ) zoo[i]->dir = STOP;
			else if ( zoo[i]->dir == DOWN ) toggle_grow(zoo[i]);

			else
			{
				ptr.y = zoo[i]->head->y + 1;
				if ( ptr.y == LINES ) ptr.y = 0;

				/* toggle growth if trying to go back */
				if ( zoo[i]->head->next && zoo[i]->head->next->y == ptr.y )
					toggle_grow(zoo[i]);

				/* change to right direction */
				else zoo[i]->dir = DOWN;
			}

			continue;

		}

	}

	return 1;
}

/* eat snake's tail */
void eat_tail ( snake *s )
{
 	/* chain pointer */
	chain *ptr;

	/* if snake disappeared, finish it */
	if( s->tail == s->head )
	{
		kill_snake(s);
		return;
	}

	/* delete the old tail before its coordinates change */
	mvaddch(s->tail->y, s->tail->x, ' ');

	/* find the chain before the tail */
	for ( ptr=s->head; ptr->next != s->tail; ptr=ptr->next );

	/* make it the new tail pointer */
	s->tail = ptr; 

	/* free memory of the old tail */
	free(s->tail->next); 

	/* set the new tail pointer to NULL */
	s->tail->next = NULL;
}


/* move head returning 0 when hitting something */
void move_head ( snake *s )
{
	chain *newhead;

	/* allocate memory for new chain */
	newhead = (chain *) malloc(sizeof(chain));
	
	/* set new head coordinates */
	switch ( s->dir )
	{

		case LEFT:
			/* move left */
			newhead->x = s->head->x - 1;
			newhead->y = s->head->y;
			
			/* get wrapped */
			if ( newhead->x < 1 ) newhead->x = COLS - 2;

			break;

		case UP:
			/* move up */
			newhead->x = s->head->x;
			newhead->y = s->head->y - 1;
			
			/* get wrapped */
			if ( newhead->y < 1 ) newhead->y = LINES - 2;
			
			break;

		case RIGHT:
			/* move right */
			newhead->x = s->head->x + 1;
			newhead->y = s->head->y;
			
			/* get wrapped */
			if ( newhead->x == COLS-1 ) newhead->x = 1;
			
			break;

		case DOWN:
			/* move down */
			newhead->x = s->head->x;
			newhead->y = s->head->y + 1;
			
			/* get wrapped */
			if ( newhead->y == LINES-1 ) newhead->y = 1;
			
			break;

	}
	
	/* set the new snake's head */
	newhead->next = s->head;
	s->head = newhead;

	/* check if snake have hitten something */
	if ( ( mvinch(s->head->y, s->head->x) & A_REVERSE ) == A_REVERSE )
		kill_snake(s);

	/* print the new head */
	else mvaddch(newhead->y, newhead->x, s->look);
}

void set_keys ( snake *ptr, int l, int u, int r, int d, int s, int g )
{
	/* set snake's control keys */
	ptr->kl = l;   /* left   */
	ptr->ku = u;   /* up     */
	ptr->kr = r;   /* right  */
	ptr->kd = d;   /* down   */
	ptr->ks = s;   /* stop   */
	ptr->kg = g;   /* go     */
}


/* kill a snake */
kill_snake ( snake *s )
{
	chain *ptr;
	
	ptr=s->head;
	while ( ptr )
	{
		chain *oldptr;

		oldptr = ptr;
		mvaddch(ptr->y, ptr->x, '.');
		ptr=ptr->next;
		free(oldptr);
	}

	/* stop its run */
	s->dir = STOP;

	/* grow it ( actually avoid shrinking it ) */
	s->grow = TRUE;

	/* mark it killed */
	s->died = TRUE;

	beep(); /* beeeep */
}

/* toggle growing state */
void toggle_grow ( snake *s )
{
	s->grow = !s->grow;
}

void draw_snake ( snake *s )
{
	chain *ptr;
	for ( ptr=s->head; ptr != NULL; ptr=ptr->next )
		mvaddch(ptr->y, ptr->x, s->look);
}

void draw_motion (void)
{
	if ( ++spriteidx == NSPRITE ) spriteidx = 0;
	mvaddch(0,0,sprite[spriteidx]);
}
